# -*- coding: utf-8 -*-
#######################################################################################
#                                                                                     #
#   This file is part of the updater4pyi Project.                                     #
#                                                                                     #
#   Copyright (C) 2013, Philippe Faist                                                #
#   philippe.faist@bluewin.ch                                                         #
#   All rights reserved.                                                              #
#                                                                                     #
#   Redistribution and use in source and binary forms, with or without                #
#   modification, are permitted provided that the following conditions are met:       #
#                                                                                     #
#   1. Redistributions of source code must retain the above copyright notice, this    #
#      list of conditions and the following disclaimer.                               #
#   2. Redistributions in binary form must reproduce the above copyright notice,      #
#      this list of conditions and the following disclaimer in the documentation      #
#      and/or other materials provided with the distribution.                         #
#                                                                                     #
#   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND   #
#   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED     #
#   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE            #
#   DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR   #
#   ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES    #
#   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;      #
#   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND       #
#   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT        #
#   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS     #
#   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                      #
#                                                                                     #
#######################################################################################

import sys
import os
import os.path
import re
import subprocess
import logging
import inspect
import urllib
import datetime

logger = logging.getLogger('updater4pyi')


# -------------------------------------


def ignore_exc(f, exc=(Exception,), value_if_exc=None):
    if (not isinstance(exc, tuple)):
        exc = (exc,);
    try:
        return f()
    except exc:
        return value_if_exc


# -------------------------------------






_TIMEDELTA_RX = r'''(?xi)
    (?P<num>\d+)\s*
    (?P<unit>
        y(ears?)?|
        mon(ths?)?|
        weeks?|
        days?|
        hours?|
        min(utes?)?|
        s(ec(onds?)?)?
    )
    (,\s*)?
    ''';
_timedelta_units = {'y':    datetime.timedelta(days=365,seconds=0, microseconds=0),
                    'mon':  datetime.timedelta(days=30, seconds=0, microseconds=0),
                    'week': datetime.timedelta(days=7,  seconds=0, microseconds=0),
                    'day':  datetime.timedelta(days=1,  seconds=0, microseconds=0),
                    'hour': datetime.timedelta(days=0,  seconds=3600, microseconds=0),
                    'min':  datetime.timedelta(days=0,  seconds=60, microseconds=0),
                    's':    datetime.timedelta(days=0,  seconds=1, microseconds=0),
                    }

def ensure_timedelta(x):
    if isinstance(x, datetime.timedelta):
        return x
    
    if isinstance(x, basestring):
        val = datetime.timedelta(0)
        for m in re.finditer(_TIMEDELTA_RX, x):
            thisvallst = [ v for k,v in _timedelta_units.iteritems()
                           if k.lower().startswith(m.group('unit')) ]
            if not thisvallst: raise ValueError("Unexpected unit: %s" %(m.group('unit')))
            val += int(m.group('num')) * thisvallst[0]
        return val

    try:
        sec = int(x)
        musec = (x-int(x))*1e6
        return datetime.timedelta(0, sec, musec)
    except ValueError:
        pass
    
    raise ValueError("Unable to parse timedelta representation: %r" %(x))


def ensure_datetime(x):
    if isinstance(x, datetime.datetime):
        return x

    if isinstance(x, basestring):
        try:
            import dateutil.parser
            return dateutil.parser.parse(x)
        except ImportError:
            pass

        for fmt in ('%Y-%m-%dT%H:%M:%S.%f',
                    '%Y-%m-%dT%H:%M:%S',
                    ):
            try:
                return datetime.strptime(x, fmt)
            except ValueError:
                pass
        raise ValueError("Can't parse date/time : %s" %(x))

    raise ValueError("Can't parse date/time: unknown type: %r" %(x))
        



# ------------------------------------

# platform utils


def is_macosx():
    return sys.platform.startswith('darwin')

def is_win():
    return sys.platform.startswith('win')

def is_linux():
    return sys.platform.startswith('linux')

def simple_platform():
    if is_macosx():
        return 'macosx'
    elif is_win():
        return 'win'
    elif is_linux():
        return 'linux'
    else:
        return sys.platform





# other utilities


def bash_quote(x):
    return "'" + x.replace("'", "'\\''") + "'"
def winshell_quote(x):
    # wait a minute, how are win/dos batch escapes ??!?
    # ... see http://technet.microsoft.com/en-us/library/cc723564.aspx
    return '"' + x.replace('"', '""') + '"'
def applescript_quote(x):
    return '"' + re.sub(r'([\\"])', r'\\\1', x) + '"'


# ------------------------------------------------------------------------

# utility to get resource files bundled with pyinstaller

def resource_path(relative_path):
    """
    Get absolute path to resource, works for dev and for PyInstaller
    """
    def base_path():
        try:
            return sys._MEIPASS
        except AttributeError:
            pass
        mainfn = inspect.stack()[-1][1]
        return os.path.abspath(os.path.dirname(mainfn))

    return os.path.join(base_path(), relative_path)


# ------------


def path2url(p):

    x = p;
    if os.sep != '/':
        # on windows: a\b -> a/b
        x = x.replace(os.sep, '/');
        
    x = urllib.pathname2url(x)
    if not x.startswith('///'):
        x = "//"+os.path.abspath(x)
    return 'file:'+x
    

# ------------


def locationIsWritable(path):
    if (os.path.isdir(path)):
        return dirIsWritable(path)
    if (os.path.isfile(path)):
        return fileIsWritable(path)
    logger.warning("location does not exist: %s", path);
    return False


def fileIsWritable(fn):
    if not is_win():
        return os.access(fn, os.W_OK)

    # otherwise, don't try to open the file itself, as it might be the executable itself
    # which is locked for write access while the program executs. Instead, test writability
    # of the containing directory.

    return dirIsWritable(os.path.dirname(fn))
    

# solution adapted from http://stackoverflow.com/a/8620444/1694896
def dirIsWritable(directory):
    if not is_win():
        return os.access(directory, os.W_OK)
    
    try:
        tmp_prefix = "upd4pyi_tmp_write_tester";
        count = 0
        filename = os.path.join(directory, tmp_prefix)
        while(os.path.exists(filename)):
            filename = "%s.%s.tmp" % (os.path.join(directory, tmp_prefix), count)
            count = count + 1
        with open(filename,"w") as f:
            pass
        os.remove(filename)
        return True
    except IOError as e:
        #print "{}".format(e)
        return False

##     # detect if admin rights are needed.
##     if not util.is_win():
##         needs_sudo = not (os.access(filetoupdate.fn, os.W_OK) and
##                           os.access(os.path.dirname(filetoupdate.fn), os.W_OK))
##     else:
##         # NOTE: This does not work under linux, because we get eg. the error
##         #       IOError: [Errno 26] Text file busy: '/home/..../dist/testpycmdline/testpycmdline'


##         # NOTE: BUG DOESN'T WORK UNDER WINDOWS, EITHER
##         ....................................
        
##         logger.debug("determining whether we need sudo: try opening %s as 'r+'.", filetoupdate.executable);
##         needs_sudo = True
##         try:
##             with open(filetoupdate.executable, 'r+') as f:
##                 pass
##             logger.debug("no sudo needed.");
##             needs_sudo = False
##         except IOError:
##             # the file is write-protected
##             logger.debug("file is write-protected: sudo will be needed.");
##             needs_sudo = True




# ------------


def run_as_admin(argv):
    cmd = [];
    if is_macosx():
        cmd = ["osascript",
               "-e",
               "do shell script " + applescript_quote(
                   " ".join([bash_quote(x) for x in argv])
                   ) + " with administrator privileges"
               ]
    elif is_linux():
        # repeated use of which() is ok because it caches result
        if which('pkexec'):
            cmd = [which('pkexec')] + argv
        elif os.environ.get('DISPLAY') and which('gksudo'):
            cmd = [which('gksudo')] + argv
        elif os.environ.get('DISPLAY') and which('kdesudo'):
            cmd = [which('kdesudo')] + argv
        elif os.environ.get('DISPLAY') and which('xterm'):
            cmd = [which('xterm'), '-e', 'sudo'] + argv
        else:
            cmd = [which('sudo')] + argv
    elif is_win():
        # our helper function
        return _run_as_admin_win(argv);
    else:
        logger.error("Platform not recognized for running process as admin: %s",
                     simple_platform())
        raise NotImplementedError

    logger.debug("Running command %r", cmd)

    # run the prepared command
    retcode = subprocess.call(cmd, stdin=None, stdout=None, stderr=None,
                              shell=False)
    if (retcode != 0):
        # produce a warning in case of an error.
        logger.warning("admin subprocess %s failed", (argv[0] if argv else None))

    return retcode


# -----------------------------------------------------------------------------------
#
# code inspired from http://stackoverflow.com/a/19719292/1694896
#
def run_win(argv, needs_sudo=False, wait=True, cwd=None):
    """
    Run a process on windows.

    Returns: the exit code of the process if `wait` is `True`, or the PID of the running
    process if `wait` is `False`.
    """

    if os.name != 'nt':
        raise RuntimeError, "This function is only implemented on Windows."

    import win32api, win32con, win32event, win32process
    from win32com.shell.shell import ShellExecuteEx
    from win32com.shell import shellcon

    # XXX TODO: isn't there a function or something we can call to massage command line params?
    cmd = winshell_quote(argv[0])
    params = " ".join([winshell_quote(x) for x in argv[1:]])
    
    cmdDir = ''
    showCmd = win32con.SW_SHOWNORMAL
    #showCmd = win32con.SW_HIDE

    if needs_sudo:
        lpVerb = 'runas'  # causes UAC elevation prompt.
    else:
        lpVerb = 'open' # just open the application

    logger.debug("running %s %s", cmd, params)

    # ShellExecute() doesn't seem to allow us to fetch the PID or handle
    # of the process, so we can't get anything useful from it. Therefore
    # the more complex ShellExecuteEx() must be used.

    # procHandle = win32api.ShellExecute(0, lpVerb, cmd, params, cmdDir, showCmd)

    optional_args = {}
    if (cwd is not None):
        optional_args['lpDirectory'] = cwd

    procInfo = ShellExecuteEx(nShow=showCmd,
                              fMask=shellcon.SEE_MASK_NOCLOSEPROCESS,
                              lpVerb=lpVerb,
                              lpFile=cmd,
                              lpParameters=params,
                              **optional_args
                              )

    if wait:
        procHandle = procInfo['hProcess']    
        obj = win32event.WaitForSingleObject(procHandle, win32event.INFINITE)
        rc = win32process.GetExitCodeProcess(procHandle)
        logger.debug("Process handle %s returned code %s", (procHandle, rc))
    else:
        # get PID
        rc = win32process.GetProcessId(procInfo['hProcess']);

    return rc








# --------------------------------------------------------------------------------------------

# Code for parse_version() taken from setuptools project,
#
# https://bitbucket.org/pypa/setuptools/src/353a4270074435faa7daa2aa0ee480e22e505f53/pkg_resources.py?at=default
#
# TODO: didn't find license information. Is there any?
#

_component_re = re.compile(r'(\d+ | [a-z]+ | \.| -)', re.VERBOSE)
_replace = {'pre':'c', 'preview':'c','-':'final-','rc':'c','dev':'@'}.get

def _parse_version_parts(s):
    for part in _component_re.split(s):
        part = _replace(part,part)
        if not part or part=='.':
            continue
        if part[:1] in '0123456789':
            yield part.zfill(8)    # pad for numeric comparison
        else:
            yield '*'+part

    yield '*final'  # ensure that alpha/beta/candidate are before final

def parse_version(s):
    """Convert a version string to a chronologically-sortable key

    This is a rough cross between distutils' StrictVersion and LooseVersion;
    if you give it versions that would work with StrictVersion, then it behaves
    the same; otherwise it acts like a slightly-smarter LooseVersion. It is
    *possible* to create pathological version coding schemes that will fool
    this parser, but they should be very rare in practice.

    The returned value will be a tuple of strings.  Numeric portions of the
    version are padded to 8 digits so they will compare numerically, but
    without relying on how numbers compare relative to strings.  Dots are
    dropped, but dashes are retained.  Trailing zeros between alpha segments
    or dashes are suppressed, so that e.g. "2.4.0" is considered the same as
    "2.4". Alphanumeric parts are lower-cased.

    The algorithm assumes that strings like "-" and any alpha string that
    alphabetically follows "final"  represents a "patch level".  So, "2.4-1"
    is assumed to be a branch or patch of "2.4", and therefore "2.4.1" is
    considered newer than "2.4-1", which in turn is newer than "2.4".

    Strings like "a", "b", "c", "alpha", "beta", "candidate" and so on (that
    come before "final" alphabetically) are assumed to be pre-release versions,
    so that the version "2.4" is considered newer than "2.4a1".

    Finally, to handle miscellaneous cases, the strings "pre", "preview", and
    "rc" are treated as if they were "c", i.e. as though they were release
    candidates, and therefore are not as new as a version string that does not
    contain them, and "dev" is replaced with an '@' so that it sorts lower than
    than any other pre-release tag.
    """
    parts = []
    for part in _parse_version_parts(s.lower()):
        if part.startswith('*'):
            if part<'*final':   # remove '-' before a prerelease tag
                while parts and parts[-1]=='*final-': parts.pop()
            # remove trailing zeros from each series of numeric parts
            while parts and parts[-1]=='00000000':
                parts.pop()
        parts.append(part)
    return tuple(parts)





# --------------------------------------------------------------------------------------------

# function which() is based on code from
# http://twistedmatrix.com/trac/browser/tags/releases/twisted-8.2.0/twisted/python/procutils.py
#
# license notice:
#
# Copyright (c) 2001-2008
# Allen Short
# Andrew Bennetts
# Apple Computer, Inc.
# Benjamin Bruheim
# Bob Ippolito
# Canonical Limited
# Christopher Armstrong
# David Reid
# Donovan Preston
# Eric Mangold
# Itamar Shtull-Trauring
# James Knight
# Jason A. Mobarak
# Jean-Paul Calderone
# Jonathan Lange
# Jonathan D. Simms
# Jürgen Hermann
# Kevin Turner
# Mary Gardiner
# Matthew Lefkowitz
# Massachusetts Institute of Technology
# Moshe Zadka
# Paul Swartz
# Pavel Pergamenshchik
# Ralph Meijer
# Sean Riley
# Software Freedom Conservancy
# Travis B. Hartwell
# Thomas Herve
# Eyal Lotem
# Antoine Pitrou
# Andy Gayton
# 
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
# 
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

# ### PhF: cache which() results
_which_cache = {}
_which_cache_first = {}

def which_clear_cache():
    _which_cache.clear()
    _which_cache_first.clear()

def which(name, flags=os.X_OK, usecache=True, firstresult=True):
    """Search PATH for executable files with the given name.

    On newer versions of MS-Windows, the PATHEXT environment variable will be
    set to the list of file extensions for files considered executable. This
    will normally include things like ".EXE". This fuction will also find files
    with the given name ending with any of these extensions.

    On MS-Windows the only flag that has any meaning is os.F_OK. Any other
    flags will be ignored.

    @type name: C{str}
    @param name: The name for which to search.

    @type flags: C{int}
    @param flags: Arguments to L{os.access}.

    @rtype: C{list}
    @param: A list of the full paths to files found, in the
    order in which they were found.
    """

    # PhF: check in the cache first, see if we've found `name` already
    if (usecache):
        if firstresult:
            if name in _which_cache_first:
                return _which_cache_first.get(name)
        else:
            if name in _which_cache:
                return _which_cache.get(name)

    result = []
    exts = filter(None, os.environ.get('PATHEXT', '').split(os.pathsep))
    path = os.environ.get('PATH', None)
    if path is None:
        return []
    for p in os.environ.get('PATH', '').split(os.pathsep):
        p = os.path.join(p, name)
        if os.access(p, flags):
            result.append(p)
        for e in exts:
            pext = p + e
            if os.access(pext, flags):
                result.append(pext)
        if (firstresult and result):
            # we've found a result, return it
            _which_cache_first[name] = result[0]
            return result[0]
        
    if usecache and result:
        # we know that firstresult=False because otherwise we would have
        # returned already.
        _which_cache[name] = result

    return result




# hook for pyinstaller

import updater4pyi
import updater4pyi.util
import os.path

def locpath(x):
    return os.path.realpath(os.path.join(os.path.dirname(updater4pyi.__file__), x))

datas = [
    (locpath('cacert.pem'), 'updater4pyi'),
    ]

if updater4pyi.util.is_linux() or updater4pyi.util.is_macosx():
    datas += [
        (locpath('installers/unix/do_install.sh'), 'updater4pyi/installers/unix'),
        ]
elif updater4pyi.util.is_win():
    datas += [
        (locpath('installers/win/do_install.exe.zip'), 'updater4pyi/installers/win'),
        ]

#from hookutils import collect_data_files
#datas = collect_data_files('updater4pyi')
print "DATAS IS\n\t%r"%(datas)



# -*- coding: utf-8 -*-
#######################################################################################
#                                                                                     #
#   This file is part of the updater4pyi Project.                                     #
#                                                                                     #
#   Copyright (C) 2013, Philippe Faist                                                #
#   philippe.faist@bluewin.ch                                                         #
#   All rights reserved.                                                              #
#                                                                                     #
#   Redistribution and use in source and binary forms, with or without                #
#   modification, are permitted provided that the following conditions are met:       #
#                                                                                     #
#   1. Redistributions of source code must retain the above copyright notice, this    #
#      list of conditions and the following disclaimer.                               #
#   2. Redistributions in binary form must reproduce the above copyright notice,      #
#      this list of conditions and the following disclaimer in the documentation      #
#      and/or other materials provided with the distribution.                         #
#                                                                                     #
#   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND   #
#   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED     #
#   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE            #
#   DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR   #
#   ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES    #
#   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;      #
#   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND       #
#   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT        #
#   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS     #
#   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                      #
#                                                                                     #
#######################################################################################

import re
import sys
import os
import os.path
import datetime

from . import util
from .upd_defs import Updater4PyiError
from .upd_log import logger



class UpdateInterface(object):
    def __init__(self, updater, progname=None, **kwargs):
        self.updater = updater
        self.progname = progname
        super(UpdateInterface, self).__init__(**kwargs)

    def start(self, **kwargs):
        """
        Start being aware of wanting to check for updates. It is up to the interface
        to decide when to check for updates, how often, etc. For example, a console
        interface would check right away, while a GUI might first load the application,
        and set a timer to check later, so that startup is not slowed down by the update
        check.
        """
        raise NotImplementedError




# -----------




class UpdateConsoleInterface(UpdateInterface):
    def __init__(self, updater, ask_before_checking=False, **kwargs):
        super(UpdateConsoleInterface, self).__init__(updater=updater, **kwargs)
        self.ask_before_checking = ask_before_checking


    def start(self):

        try:
            
            self._runupdatecheck();
            
        except Updater4PyiError as e:
            print "\n"
            print "------------------------------------------------------------"
            print "Error: %s" % (e)
            print "Software update aborted."
            print "------------------------------------------------------------"
            print "\n"
            return

        # return to the main program.
        return


    def _runupdatecheck(self):
        #
        # See if we should ask before checking.
        #
        if (self.ask_before_checking):
            if (not self._ynprompt("Do you wish to check for software updates? (y/n) ")):
                return

        #
        # Check for updates.
        #
        upd_info = self.updater.check_for_updates()

        if (upd_info is None):
            # no updates.
            print "No updates available."
            return

        #
        # There's an update, prompt the user.
        #
        print ""
        print "-----------------------------------------------------------"
        print ""
        print ("A new software update is available (%sversion %s)"
               % (self.progname+" " if self.progname else "", upd_info.version))
        print ""

        if (self._ynprompt("Do you want to install it? (y/n) ")):
            #
            # yes, install update
            #
            self.updater.install_update(upd_info)
            #
            # update installed.
            #
            print ""
            print "Update installed. Quitting. Please restart the program."
            print ""
            print "-----------------------------------------------------------"
            print ""
            sys.exit(0)
        else:
            print ""
            print "Not installing update."
            print ""
            print "-----------------------------------------------------------"
            print ""
        
        
    def _ynprompt(self, msg):
        yn = raw_input(msg)
        return re.match(r'\s*y(es)?\s*', yn, re.IGNORECASE) is not None




# ---------------------------------------------------------------------


# initial check by default 1 minute after app startup, so as not to slow down app startup.
DEFAULT_INIT_CHECK_DELAY = datetime.timedelta(days=0, seconds=60, microseconds=0)
# subsequent checks every week by default
DEFAULT_CHECK_INTERVAL = datetime.timedelta(days=7, seconds=0, microseconds=0)


_SETTINGS_ALL = ['check_for_updates_enabled', 'init_check_delay', 'check_interval',
                 'last_check']

class UpdateGenericGuiInterface(UpdateInterface):
    def __init__(self, updater, ask_before_checking=True, **kwargs):
        super(UpdateGenericGuiInterface, self).__init__(updater, **kwargs)

        self.ask_before_checking = ask_before_checking;
        
        self.update_installed = False
        self.is_initial_delay = True

        self.is_currently_checking = False

        # load settings
        d = self.load_settings(self._settings_all_keys())
        self.init_check_delay = util.ensure_timedelta(d.get('init_check_delay', DEFAULT_INIT_CHECK_DELAY))
        self.check_interval = util.ensure_timedelta(d.get('check_interval', DEFAULT_CHECK_INTERVAL))
        self.check_for_updates_enabled = d.get('check_for_updates_enabled', True)
        self.last_check = util.ensure_datetime(d.get('last_check', datetime.datetime(1970, 1, 1)))
        if (self.ask_before_checking):
            self.asked_before_checking = d.get('asked_before_checking', False);
        
    
    def start(self):
        logger.debug("Starting interface (generic gui)")
        self.schedule_next_update_check()


    # properties

    def initCheckDelay(self):
        return self.init_check_delay

    def setInitCheckDelay(self, init_check_delay, save=True):
        self.init_check_delay = util.ensure_timedelta(init_check_delay)
        if save: self.save_settings({'init_check_delay': self.init_check_delay})

    def checkInterval(self):
        return self.check_interval

    def setCheckInterval(self, check_interval, save=True):
        self.check_interval = util.ensure_timedelta(check_interval)
        if save: self.save_settings({'check_interval': self.check_interval})

    def checkForUpdatesEnabled(self):
        return self.check_for_updates_enabled

    def setCheckForUpdatesEnabled(self, enabled, save=True):
        self.check_for_updates_enabled = enabled
        # save setting to settings file
        if save: self.save_settings({'check_for_updates_enabled': self.check_for_updates_enabled})
        # also, schedule the next update check
        self.schedule_next_update_check()

    def lastCheck(self):
        return self.last_check

    def setLastCheck(self, last_check, save=True):
        self.last_check = util.ensure_datetime(last_check)
        if save: self.save_settings({'last_check': self.last_check})

    # ------------

    def all_settings(self):
        """
        Utility to get all settings. Useful for subclasses; this doesn't need to be reimplemented.
        """
        return dict([(k,getattr(self,k)) for k in self._settings_all_keys()])


    def _settings_all_keys(self):
        """
        Returns a list of relevant settings keys for this object. Includes `'aksed_before_checking'` only
        if the `ask_before_checking` argument given to the constructor was `True`.
        """
        return _SETTINGS_ALL + (['asked_before_checking'] if self.ask_before_checking else [])

    # ----------------------------------------------

    def check_for_updates(self):
        """
        Perform a possible update check. You don't have to reimplement this function, the default
        implementation should be good enough and relies on your implementations of `ask_to_update()`
        and `ask_to_restart()`.

        If the update check isn't due yet, this function does not do the update check. If
        you want to force an update check, call `do_check_for_updates()`.
        """
        logger.debug("UpdateGenericGuiInterface: check_for_updates()")

        if (self.update_installed):
            logger.warning("We have already installed an update and pending restart.")
            return

        logger.debug("self.is_initial_delay=%r, self.timedelta_remaining_to_next_check()=%r",
                     self.is_initial_delay, self.timedelta_remaining_to_next_check())

        # if we were called just after the initial delay, reset this flag.
        if (self.is_initial_delay):
            self.is_initial_delay = False

        # now, really check that we are due for software update check.
        #   * we might be after an initial delay, and while the app was just started, updates are checked
        #     for eg. once a month and a check is not yet due, yet this function is called by timeout.
        #   * even if we are not at the initial delay, the user may have disabled update checks in the
        #     settings between the scheduling of the update check and now.
        if (not self.is_check_now_due()):
            # software update check is not yet due (not even in the next 10 seconds). Just
            # schedule next check.
            logger.debug("Update check is not yet due. Postpone.")
            self.schedule_next_update_check()
            return

        try:
            self.do_check_for_updates()
        finally:
            self.schedule_next_update_check()


    def do_check_for_updates(self):
        """
        Actually perform the udpate check. Call this function if you want to force an
        update check even though it's not yet due. If you want to periodically possibly
        check only if a check is due, then call `check_for_updates()` instead.

        Returns:
            - `None` if we asked the user for the first time if they want to check
              regularly for updates, and they refused.
            - `False` if no new update is available
            - a tuple if a new update is available:
                - `(True, rel_info)` if the user installed the update but did not
                  restart the app;
                - `(False, rel_info)` if the user declined to install the update now
            - the tuple `(False, None, error_str)` if an error occurred while checking
              for updates.
        """
        if self.is_currently_checking:
            return None
        try:
            self.is_currently_checking = True

            # check for updates

            if (self.ask_before_checking and not self.asked_before_checking):
                # ask before we check.
                if (self.ask_first_time() != True):
                    logger.debug("UpdateGenericGuiInterface: are told not to check for updates.");
                    self.setCheckForUpdatesEnabled(False);
                    self.asked_before_checking = True
                    self.save_settings({'asked_before_checking': True})
                    return None

            rel_info = self.updater.check_for_updates()

            if (rel_info is None):
                # no updates.
                logger.debug("UpdateGenericGuiInterface: No updates available.")
                return False

            logger.debug("Update (version %s) is available.", rel_info.get_version())

            #
            # There's an update, prompt the user.
            #
            if self.ask_to_update(rel_info):
                #
                # yes, install update
                #
                self.updater.install_update(rel_info)
                self.update_installed = True
                #
                # update installed.
                #
                if self.ask_to_restart():
                    self.updater.restart_app()
                    return (True, rel_info) # whatever, our app will have exited anyway

                # return to the main program.
                return (True, rel_info)

            logger.debug("UpdateGenericGuiInterface: Not installing update.")
            # return to the main program.
            return (False, rel_info)
        
        except Updater4PyiError as e:
            logger.warning("Error while checking for updates: %s", e)
            return (False, None, unicode(e))
            
        finally:
            self.last_check = datetime.datetime.now()
            self.save_settings({'last_check': self.last_check})
            self.is_currently_checking = False


    def is_check_now_due(self, tolerance=datetime.timedelta(days=0, seconds=10)):
        return (self.check_for_updates_enabled and
                self.timedelta_remaining_to_next_check() <= tolerance)

    def schedule_next_update_check(self):
        if not self.check_for_updates_enabled:
            logger.debug("UpdateGenericGuiInterface:Not scheduling update check because we were "
                         "asked not to check for updates.")
            return

        if (self.is_initial_delay):
            self.set_timeout_check(self.init_check_delay)
            logger.debug("UpdateGenericGuiInterface: requested initial single-shot timer for %r seconds"
                         %(self.init_check_delay))
        else:
            timedelta_remaining = self.timedelta_remaining_to_next_check()
            if (timedelta_remaining <= datetime.timedelta(0)):
                logger.debug("UpdateGenericGuiInterface: software update check due now already, checking")
                self.check_for_updates()
                return
            self.set_timeout_check(timedelta_remaining)
            logger.debug("UpdateGenericGuiInterface: requested single-shot timer for %r",
                         timedelta_remaining)


    def timedelta_remaining_to_next_check(self):
        return ((self.last_check + self.check_interval) - datetime.datetime.now())



    # ------------------------------------------------------------------------------
    # the following methods need to be reimplemented, using the gui toolkit at hand.
    # ------------------------------------------------------------------------------
    

    def ask_first_time(self):
        """
        Subclasses should prompt the user whether they want to regularly look for updates.

        This is prompted to the user only if the main program set `ask_before_checking` to `True` in the
        constructor of this object.
        
        Return TRUE if the program should regularly check for updates, or FALSE if not.
        """
        raise NotImplementedError
        
    def ask_to_update(self, rel_info):
        """
        Subclasses should prompt the user whether they want to install the update `rel_info` or not.
        
        Note: Interfaces may also present additional buttons such as "Never check for updates", or
        "Skip this update", and set properties and/or settings accordingly with e.g.
        `setCheckForUpdatesEnabled()`.
        
        Return TRUE if the program should be restarted, or FALSE if not.
        """
        raise NotImplementedError
        
    def ask_to_restart(self):
        """
        Subclasses should prompt the user to restart the program after a successful update.

        Return TRUE if the program should be restarted, or FALSE if not.
        """
        raise NotImplementedError
        
    def set_timeout_check(self, interval_timedelta):
        """
        Subclasses should reimplement this function to call the function `check_for_updates()` after
        `interval_timedelta`. `interval_timedelta` is a `datetime.timedelta` object.
        """
        raise NotImplementedError

    def load_settings(self, keylist):
        """
        Subclasses may reimplement this function to cusomize where and how the settings are stored,
        usually using a toolbox-specific utility, such as QSettings in PyQt4.
        """
        raise NotImplementedError

    def save_settings(self, d=None):
        """
        Save the given settings in the dictionary `d` to some local settings. If d is None, then
        all settings should be saved, effectively taking `d` to be the dictionary returned by
        `all_settings()`.
        """
        raise NotImplementedError






# -*- coding: utf-8 -*-
#######################################################################################
#                                                                                     #
#   This file is part of the updater4pyi Project.                                     #
#                                                                                     #
#   Copyright (C) 2013, Philippe Faist                                                #
#   philippe.faist@bluewin.ch                                                         #
#   All rights reserved.                                                              #
#                                                                                     #
#   Redistribution and use in source and binary forms, with or without                #
#   modification, are permitted provided that the following conditions are met:       #
#                                                                                     #
#   1. Redistributions of source code must retain the above copyright notice, this    #
#      list of conditions and the following disclaimer.                               #
#   2. Redistributions in binary form must reproduce the above copyright notice,      #
#      this list of conditions and the following disclaimer in the documentation      #
#      and/or other materials provided with the distribution.                         #
#                                                                                     #
#   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND   #
#   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED     #
#   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE            #
#   DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR   #
#   ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES    #
#   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;      #
#   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND       #
#   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT        #
#   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS     #
#   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                      #
#                                                                                     #
#######################################################################################


import re
import sys
import inspect
import os
import os.path
import stat
import collections
import zipfile
import tarfile
import json
import glob
import subprocess
import tempfile
import shutil

from . import util
from . import upd_version
from .upd_log import logger
from .upd_defs import RELTYPE_UNKNOWN, RELTYPE_EXE, RELTYPE_ARCHIVE, RELTYPE_BUNDLE_ARCHIVE
from .upd_defs import Updater4PyiError
import upd_downloader


# --------------------------------


FileToUpdate = collections.namedtuple('FileToUpdate', ('fn', 'reltype', 'executable',));


def determine_file_to_update():
    """
    Returns a FileToUpdate(fn=.., reltype=.., executable=...) named tuple.
    """
    
    executable = sys.executable
    updatefile = os.path.realpath(sys.executable)
    reltype = None

    logger.debug("trying to determine pyi executable to update. sys.executable=%s; sys._MEIPASS=%s",
                 sys.executable, (sys._MEIPASS if hasattr(sys, '_MEIPASS') else '<no sys._MEIPASS>'))

    if (sys.platform.startswith('darwin')):
        # see if we are a Mac OS X bundle
        (alllastdir,fn) = os.path.split(sys.executable);
        (allbeforelastdir,lastdir) = os.path.split(alllastdir);
        (allbeforebeforelastdir,beforelastdir) = os.path.split(allbeforelastdir);

        logger.debug("platform is Mac OS X; alllastdir=%s, beforelastdir=%s, lastdir=%s, fn=%s",
                     alllastdir, beforelastdir, lastdir, fn)

        if (lastdir == 'MacOS' and beforelastdir == 'Contents'):
            # we're in a Mac OS X bundle, so the actual "executable" should point to the .app file
            reltype = RELTYPE_BUNDLE_ARCHIVE
            updatefile = allbeforebeforelastdir
            logger.debug("We're a bundle: updatefile=%s", updatefile)

    if reltype is None:
        # if we're not already a bundle, check whether we're a directory to update
        if (hasattr(sys, '_MEIPASS')):
            meipass = os.path.realpath(sys._MEIPASS);
            if (updatefile.startswith(meipass)):
                # pyinstaller files are installed directly: it's the dir we need to update
                reltype = RELTYPE_ARCHIVE
                updatefile = meipass;

    if reltype is None:
        # otherwise, we're a self-contained executable.
        reltype = RELTYPE_EXE


    logger.debug("got FileToUpdate(fn=%r, reltype=%d, executable=%s)",
                 updatefile, reltype, executable)

    return FileToUpdate(fn=updatefile, reltype=reltype, executable=executable)





# ------------------------------------------------------------------------

_updater = None

def get_updater():
    return _updater



class Updater(object):
    def __init__(self, current_version, update_source):
        """
        Instantiates an `Updater`, with updates provided by the source `update_source` (a
        `upd_source.UpdateSource` subclass instance).

        ...........NO ..............., and the user interface
        `update_interface` (itself an `upd_iface.UpdateInterface` subclass instance).

        The `current_version` is the current version string of the software, and will
        be provided to the `update_source`.

        ...........NO................If `update_interface` is `None`, then a default console interface is provided
        (of class `upd_iface.UpdateConsoleInterface`).
        """

        # sys._MEIPASS seems to be set all the time, even we don't self-extract.
        if (not hasattr(sys, '_MEIPASS')):
            raise Updater4PyiError("This installation is not built with pyinstaller.")

        self._update_source = update_source

        logger.debug("source is %r" %(self._update_source))

        self._current_version = current_version
        self._file_to_update = determine_file_to_update()

        super(Updater, self).__init__()


    def update_source(self):
        return self._update_source

    def current_version(self):
        return self._current_version

    def file_to_update(self):
        return self._file_to_update


    # -------------------------------------------


    def check_for_updates(self):
        """
        Perform an update check.

        Queries the source for possible updates, which matches our system. If a software update
        is found, then a `upd_source.BinReleaseInfo` object is returned, describing the software
        update. Otherwise, if no update is available, `None` is returned.
        """
        
        releases = self._update_source.get_releases(newer_than_version=self._current_version)

        logger.debug("releases=%r" %(releases))

        if (releases is None):
            logger.warning("Software Update Source returned a None release list!")
            return None

        wanted_reltype = self._file_to_update.reltype

        # this is current version
        curver = util.parse_version(self._current_version)

        # select the releases that match our criteria;
        # also sort the releases by version number.
        rel_w_parsedversion = [(r, util.parse_version(r.get_version())) for r in releases]
        releases2 = sorted([(rel, relparsedver)
                            for (rel, relparsedver) in rel_w_parsedversion
                            if (rel.get_reltype() == wanted_reltype and
                                rel.get_platform() == util.simple_platform() and
                                relparsedver > curver)
                            ],
                           key=lambda r2: r2[1],
                           reverse=True);

        if (releases2):
            # found release(s) which are strictly newer, and which satisfy our format requirements.
            return releases2[0][0]

        # no update found
        logger.debug("No update found.")
        return None


    # ------------------------------------------

    SPECIAL_ZIP_FILES = ('_updater4pyi_metainf.json',
                         '_METAINF',
                         )

    class _ExtractLocation(object):
        def __init__(self, filetoupdate, needs_sudo, **kwargs):
            self.filetoupdate = filetoupdate
            self.needs_sudo = needs_sudo

            super(Updater._ExtractLocation, self).__init__(**kwargs)


        def findextractto(self, namelist):

            (basedir, basefn) = os.path.split(self.filetoupdate.fn)

            self.extractto = None
            self.installto = None
            self.extracttotemp = False

            self.extractstodir = True
            if ([True for x in namelist if not x.startswith(basefn) and x not in Updater.SPECIAL_ZIP_FILES]):
                # the zip file doesn't extract into a single dir--there are files with different prefixes.
                # so extract into a single dir ourselves.
                self.extractstodir = False
                self.extractto = self.filetoupdate.fn
                if not self.needs_sudo:
                    # create the directory only if we don't need superuser rights for installation
                    try:
                        os.mkdir(self.filetoupdate.fn)
                    except OSError as e:
                        raise Updater4PyiError("Failed to create directory %s!" %(self.filetoupdate.fn))
            else:
                self.extractto = basedir

            if self.needs_sudo:
                # in any case, if we need superuser rights for installation, then extract
                # to a temporary directory.
                self.installto = self.extractto
                self.extracttotemp = True
                # NOTE: Don't change prefix and suffix, this name template is relied upon by do_install.exe !!
                self.extractto = tempfile.mkdtemp(suffix='', prefix='upd4pyi_tmp_xtract_', dir=None)

            return self.extractto


    def install_update(self, rel_info):

        # first, save the file locally.
        tmpfile = tempfile.NamedTemporaryFile(mode='w+b', prefix='upd4pyi_tmp_', dir=None, delete=False)

        url = rel_info.get_url();

        try:
            self.download_file(url, tmpfile)
        except IOError as e: 
            if hasattr(e, 'code'): # HTTPError 
                raise Updater4PyiError('Got HTTP error: %d %s' %(e.code, e.reason))
            elif hasattr(e, 'reason'): # URLError 
                raise Updater4PyiError('Connection error: %s' %(e.reason))
            else:
                raise Updater4PyiError('Error: %s' %(str(e)))

        # TODO: add support for download verifyer (MD5/SHA or GPG signature) ??
        # ...
        # for now, this simple reimplementable function.
        #
        if (not self.verify_download(rel_info, tmpfile)):
            logger.warning("Failed to download %s : download verification failed.", url);
            os.unlink(tmpfile);
            raise Updater4PyiError("Failed to download software update: verification failed.")

        # at this point, file is downloaded and on disk.

        # the file/directory we have to update.
        filetoupdate = self._file_to_update;

        # do we need superuser access for performing the install?
        needs_sudo = not (util.locationIsWritable(filetoupdate.fn) and
                          util.dirIsWritable(os.path.dirname(filetoupdate.fn)))


        # determine if we will work in the temporary dir only and call an external utility (e.g. do_install.exe)
        # or if we will directly unpack e.g. the zip file at the right location.
        #
        # reasons for staying in temp locations are:
        #   * final location requires root access
        #   * windows: can't overwrite files of the running process.
        needs_work_in_temp_dir = needs_sudo or util.is_win();

        logger.debug("installation will need sudo? %s", needs_sudo)

        reltype_is_dir = filetoupdate.reltype in (RELTYPE_BUNDLE_ARCHIVE,
                                                  RELTYPE_ARCHIVE);

        extractedfile = None
        installto = None
        extractloc = None


        def cleanuptempfiles():
            if (tmpfile.name and os.path.exists(tmpfile.name)):
                logger.debug("cleaning up maybe %s", tmpfile.name)
                util.ignore_exc(lambda : os.unlink(tmpfile.name), OSError)

            if extractloc is not None and extractloc.extracttotemp and os.path.exists(extractloc.extractto):
                logger.debug("cleaning up maybe %s", extractloc.extractto)
                util.ignore_exc(lambda : shutil.rmtree(extractloc.extractto), OSError)


        # move that file out of the way, but keep it as backup. So just rename it.
        backupfilename = _backupname(filetoupdate.fn)
        if not needs_work_in_temp_dir:
            try:
                os.rename(filetoupdate.fn, backupfilename);
            except OSError as e:
                cleanuptempfiles();
                raise Updater4PyiError("Failed to rename file %s!" %(str(e)))



        def failure_cleanupandrestorebackup():
            # files may have been downloaded/extracted to some location

            cleanuptempfiles()

            if extractedfile and os.path.exists(extractedfile):
                logger.debug("cleaning up maybe %s", extractedfile)
                util.ignore_exc(lambda : shutil.rmtree(extractedfile), OSError)

            if needs_work_in_temp_dir:
                # no backup was generated anyway at this point
                return

            logger.debug("cleaning up maybe %s", filetoupdate.fn)
            util.ignore_exc(lambda : shutil.rmtree(filetoupdate.fn), OSError)

            try:
                logger.debug("restoring backup %s -> %s", backupfilename, filetoupdate.fn)
                shutil.move(backupfilename, filetoupdate.fn)
            except OSError as e:
                logger.error("Software Update Error: Failed to restore backup %s of %s! %s\n"
                             % (backupfilename, filetoupdate.fn, str(e)))

            return


        try:
            if (reltype_is_dir):
                # we are updating the directory itself. So make sure we download an archive file.

                extractloc = Updater._ExtractLocation(filetoupdate=filetoupdate,
                                                      needs_sudo=needs_work_in_temp_dir)

                logger.debug("extractloc: %r", extractloc.__dict__)

                if (zipfile.is_zipfile(tmpfile.name)):
                    # ZIP file
                    thezipfile = zipfile.ZipFile(tmpfile.name, 'r')

                    # extract the ZIP file to our directory.

                    extractloc.findextractto(namelist=thezipfile.namelist())
                    extractto = extractloc.extractto

                    permdata = None
                    if ('_updater4pyi_metainf.json' in thezipfile.namelist()):
                        # adjust permissions on files.
                        try:
                            permdata = json.load(thezipfile.open('_updater4pyi_metainf.json'))
                        except ValueError as e:
                            logger.warning("Invalid JSON data in metainf file _updater4pyi_metainf.json: %s" %(str(e)))

                    # iterate over files to extract with executable permissions set.
                    for zinfo in thezipfile.infolist():
                        if zinfo.filename in Updater.SPECIAL_ZIP_FILES:
                            continue
                        thezipfile.extract(zinfo, extractto)
                        os.chmod(os.path.join(extractto, zinfo.filename), 0755) # make executable
                    thezipfile.close()

                    # override some permissions with a special metainfo file.
                    if permdata and 'permissions' in permdata:
                        for (pattern,perm) in permdata['permissions'].iteritems():
                            logger.debug("pattern: %s to perms=%s" %(pattern, perm))
                            # int(s, 0) converts s to int, parsing prefixes '0' (octal), '0x' (hex)
                            # cf. http://stackoverflow.com/questions/604240/
                            iperm = int(perm,0);
                            for it in glob.iglob(os.path.join(filetoupdate.fn, pattern)):
                                logger.debug("Changing permissions of %s to %#o" %(it, iperm))
                                try:
                                    os.chmod(it, iperm)
                                except OSError:
                                    logger.warning("Failed to set permissions to file %s. Ignoring." %(it));
                                    pass

                    # remove the temporary downloaded file.
                    os.unlink(tmpfile.name)

                elif tarfile.is_tarfile(tmpfile.name):
                    # TAR[/GZ/BZIP2] file
                    thetarfile = tarfile.open(tmpfile.name, 'r');
                    # extract the ZIP file to our directory.

                    extractloc.findextractto(namelist=thetarfile.getnames())
                    extractto = extractloc.extractto

                    thezipfile.extractall(extractto)
                    thezipfile.close()

                    # remove the temporary downloaded file.
                    os.unlink(tmpfile.name)

                else:
                    raise Updater4PyiError("Downloaded file %s is not a recognized archive."
                                           %(os.path.basename(tmpfile.name)))

                # now, set installto if we need a sudo install
                extractedfile = os.path.join(extractto, os.path.basename(filetoupdate.fn))
                installto = extractloc.installto

                # also, check that we've extracted a valid archive which replaces the same file.
                fnreltoextract = os.path.relpath(filetoupdate.executable, start=filetoupdate.fn);
                if not os.path.exists(os.path.join(extractedfile, fnreltoextract)):
                    logger.error("Update package doesn't contain file %s in %s",
                                 fnreltoextract, extractedfile);
                    raise Updater4PyiError("Update package is malformed: can't find executable");

            else:
                # make sure the file is executable
                os.chmod(tmpfile.name,
                         stat.S_IREAD|stat.S_IWRITE|stat.S_IEXEC|
                         stat.S_IRUSR|stat.S_IWUSR|stat.S_IXUSR|
                         stat.S_IRGRP|stat.S_IXGRP|
                         stat.S_IROTH|stat.S_IXOTH
                         )

                if not needs_work_in_temp_dir:
                    # following docs: these may be on different filesystems, and docs specify that os.rename()
                    # may fail in that case. So use shutil.move() which should work.
                    shutil.move(tmpfile.name, filetoupdate.fn)
                else:
                    # the ready file, and the install location. This will be installed by the sudo script.
                    extractedfile = tmpfile.name
                    installto = filetoupdate.fn

            # do possibly the sudo install if needed
            if needs_work_in_temp_dir:
                if util.is_linux() or util.is_macosx():
                    res = util.run_as_admin([util.which('bash'),
                                             util.resource_path('updater4pyi/installers/unix/do_install.sh'),
                                             filetoupdate.fn, backupfilename, extractedfile, installto])
                    if (res != 0):
                        raise Updater4PyiError("Can't install the update to the final location %s!" %(installto))
                elif util.is_win():
                    # first, copy do_install.exe and its dependencies to some path out of the way, and
                    # instruct them to auto-destroy.
                    doinstalldirname = tempfile.mkdtemp(prefix='upd4pyi_tmp_')
                    doinstallzipfile = zipfile.ZipFile(
                        util.resource_path('updater4pyi/installers/win/do_install.exe.zip'),
                        'r')
                    doinstallzipfile.extractall(doinstalldirname)
                    # now, run do_install.exe
                    manage_install_cmd = [os.path.join(doinstalldirname, 'manage_install.exe'),
                                          str(os.getpid()),
                                          ('1' if needs_sudo else '0'),
                                          filetoupdate.fn,
                                          backupfilename,
                                          extractedfile,
                                          installto,
                                          doinstalldirname,
                                          filetoupdate.executable
                                       ]
                    logger.debug("Running %r as %s", manage_install_cmd, ("admin" if needs_sudo else "normal user"))
                    util.run_win(argv=manage_install_cmd,
                                 # manage_install will itself run do_install as sudo if needed. Don't run
                                 # manage_install as root, because manage_install is also responsible of
                                 # relaunching us.
                                 needs_sudo=False,
                                 wait=False,
                                 cwd=os.path.expanduser("~"), # some path out of our dir, which needs to be deleted.
                                 )
                    sys.exit(0)
                else:
                    logger.error("I don't know your platform to run sudo install on: %s", util.simple_platform())
                    raise RuntimeError("Unknown platform for sudo install: %s" %(util.simple_platform()))

        except Exception:
            logger.error("Software Update Error: %s\n" %(str(sys.exc_info()[1])));
            failure_cleanupandrestorebackup()
            raise

        # cleaning up temp files
        logger.debug("cleaning up temp files")
        cleanuptempfiles()

        # remove the backup.
        if not needs_work_in_temp_dir:
            #DEBUG: logger.warning("For debugging & possible unstability, NOT removing backup.")
            if (reltype_is_dir):
                logger.debug("removing backup directory %s", backupfilename)
                try:
                    shutil.rmtree(backupfilename)
                except (OSError,IOError):
                    logger.warning("Failed to remove backup directory %s !", backupfilename)
                    # e.g. this might happen if the executable is on some filesystems such as sshfs
            else:
                logger.debug("removing backup file %s", backupfilename)
                try:
                    os.unlink(backupfilename)
                except (OSError,IOError):
                    logger.warning("Failed to remove backup file %s !", backupfilename)



    def download_file(self, theurl, fdst):

        logger.debug("fetching URL %s to temp file %s ...", theurl, util.ignore_exc(lambda : fdst.name))

        fdata = upd_downloader.url_opener.open(theurl);
        shutil.copyfileobj(fdata, fdst)
        fdata.close()
        fdst.close()

        logger.debug("... done.")


    def verify_download(self, rel_info, tmpfile):
        # TODO: add support for GPG signing, MD5/SHA-1 checksum checks etc... ???
        # Note: might not be necessary with secure https downloads with checked certificate
        return True


    # ---------------------------------------------------------------

    # utility: restart this application

    def restart_app(self):

        exe = self._file_to_update.executable

        # the exe_cmd

        if util.is_macosx() or util.is_linux():
            exe_cmd = util.bash_quote(exe)

            if (util.is_macosx() and
                exe == self._file_to_update.executable and
                self._file_to_update.fn.lower().endswith('.app')):
                # Mac OS X, we are launching this exact executable, which is known to be a .app;
                # --> use 'open FooBar.app' instead.
                exe_cmd = 'open '+util.bash_quote(self._file_to_update.fn) # the .app

            this_pid = os.getpid()
            subprocess.Popen("while ps -p %d >/dev/null; do sleep 1; done; ( %s & )"
                             %(this_pid, exe_cmd),
                             shell=True)
            sys.exit(0)

        elif util.is_win():
            # we don't need to implement this, on windows the external process manage_install.exe
            # also takes care of restarting us.
            raise RuntimeError("Can't use restart_app on windows. The manage_install.exe process already "
                               "takes care of that.")

        else:
            logger.warning("I don't know about your platform. You'll have to restart this "
                           "program by yourself like a grown-up. I'm exiting now! Have fun.")
            sys.exit(0)


            
# --------------------------------------------------------------
    

def _backupname(filename):
    try_suffix = '.bkp'
    n = 1
    while (n < 999 and os.path.exists(filename+try_suffix)):
        try_suffix = '.'+str(n)+'.bkp'
        n += 1;
    if (os.path.exists(filename+try_suffix)):
        raise Updater4PyiError("Can't figure out a backup name for file %s!!" %(filename))
    logger.debug("Got backup name: %s" %(filename+try_suffix))
    return filename+try_suffix;








# -*- coding: utf-8 -*-
#######################################################################################
#                                                                                     #
#   This file is part of the updater4pyi Project.                                     #
#                                                                                     #
#   Copyright (C) 2013, Philippe Faist                                                #
#   philippe.faist@bluewin.ch                                                         #
#   All rights reserved.                                                              #
#                                                                                     #
#   Redistribution and use in source and binary forms, with or without                #
#   modification, are permitted provided that the following conditions are met:       #
#                                                                                     #
#   1. Redistributions of source code must retain the above copyright notice, this    #
#      list of conditions and the following disclaimer.                               #
#   2. Redistributions in binary form must reproduce the above copyright notice,      #
#      this list of conditions and the following disclaimer in the documentation      #
#      and/or other materials provided with the distribution.                         #
#                                                                                     #
#   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND   #
#   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED     #
#   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE            #
#   DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR   #
#   ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES    #
#   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;      #
#   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND       #
#   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT        #
#   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS     #
#   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                      #
#                                                                                     #
#######################################################################################


version_str = '0.4'




# -*- coding: utf-8 -*-
#######################################################################################
#                                                                                     #
#   This file is part of the updater4pyi Project.                                     #
#                                                                                     #
#   Copyright (C) 2013, Philippe Faist                                                #
#   philippe.faist@bluewin.ch                                                         #
#   All rights reserved.                                                              #
#                                                                                     #
#   Redistribution and use in source and binary forms, with or without                #
#   modification, are permitted provided that the following conditions are met:       #
#                                                                                     #
#   1. Redistributions of source code must retain the above copyright notice, this    #
#      list of conditions and the following disclaimer.                               #
#   2. Redistributions in binary form must reproduce the above copyright notice,      #
#      this list of conditions and the following disclaimer in the documentation      #
#      and/or other materials provided with the distribution.                         #
#                                                                                     #
#   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND   #
#   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED     #
#   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE            #
#   DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR   #
#   ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES    #
#   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;      #
#   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND       #
#   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT        #
#   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS     #
#   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                      #
#                                                                                     #
#######################################################################################


import logging

import httplib
import ssl
import socket
import shutil
import urllib2

from . import upd_version
from . import util
from .upd_log import logger


# -------------------------------

CERT_FILE = util.resource_path('updater4pyi/cacert.pem');#'root.crt');

class ValidHTTPSConnection(httplib.HTTPConnection):
    """
    HTTPS connection based on httplib.HTTPConnection, with certificate validation.
    """

    default_port = httplib.HTTPS_PORT

    def __init__(self, *args, **kwargs):
        httplib.HTTPConnection.__init__(self, *args, **kwargs)

    def connect(self):
        """
        Connect to a host on a given (SSL) port.
        """

        logger.debug("Connecting via HTTPS to %s:%d.", self.host, self.port)
        
        sock = socket.create_connection((self.host, self.port),
                                        self.timeout, self.source_address)
        if self._tunnel_host:
            self.sock = sock
            self._tunnel()
        self.sock = ssl.wrap_socket(sock,
                                    ca_certs=CERT_FILE,
                                    cert_reqs=ssl.CERT_REQUIRED)


class ValidHTTPSHandler(urllib2.HTTPSHandler):

    def https_open(self, req):
            return self.do_open(ValidHTTPSConnection, req)



url_opener = urllib2.build_opener(ValidHTTPSHandler)
url_opener.addheaders = [('User-agent', 'Updater4Pyi-SoftwareUpdater %s'%(upd_version.version_str))]




# -*- coding: utf-8 -*-
#######################################################################################
#                                                                                     #
#   This file is part of the updater4pyi Project.                                     #
#                                                                                     #
#   Copyright (C) 2013, Philippe Faist                                                #
#   philippe.faist@bluewin.ch                                                         #
#   All rights reserved.                                                              #
#                                                                                     #
#   Redistribution and use in source and binary forms, with or without                #
#   modification, are permitted provided that the following conditions are met:       #
#                                                                                     #
#   1. Redistributions of source code must retain the above copyright notice, this    #
#      list of conditions and the following disclaimer.                               #
#   2. Redistributions in binary form must reproduce the above copyright notice,      #
#      this list of conditions and the following disclaimer in the documentation      #
#      and/or other materials provided with the distribution.                         #
#                                                                                     #
#   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND   #
#   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED     #
#   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE            #
#   DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR   #
#   ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES    #
#   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;      #
#   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND       #
#   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT        #
#   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS     #
#   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                      #
#                                                                                     #
#######################################################################################


import sys
import re
import os
import os.path
import logging
import copy
import json
import inspect
import urllib2

from . import util
from .upd_defs import RELTYPE_UNKNOWN, RELTYPE_EXE, RELTYPE_ARCHIVE, RELTYPE_BUNDLE_ARCHIVE
from .upd_defs import Updater4PyiError
from . import upd_downloader
from .upd_log import logger


# ---------------------------------------------------------------------


class BinReleaseInfo(object):
    def __init__(self, version=None, filename=None, url=None,
                 reltype=RELTYPE_UNKNOWN,
                 platform=None,
                 **kwargs):

        if not version:
            raise ValueError("BinReleaseInfo(): version is not set!")

        self.version = version
        self.filename = filename
        self.url = url
        self.reltype = reltype
        self.platform = platform

        for k,v in kwargs.iteritems():
            setattr(self, k, v)


    def get_version(self):
        return self.version

    def get_filename(self):
        return self.filename

    def get_url(self):
        return self.url

    def get_reltype(self):
        return self.reltype

    def get_platform(self):
        return self.platform

    def __repr__(self):
        return (self.__class__.__name__+'('+
                ", ".join([ '%s=%r' % (k,v)
                            for (k,v) in self.__dict__.iteritems() ]) +
                ')')




# --------------------------------------------------------------



class UpdateSource(object):
    """
    Base abstract class for an update source.

    Subclasses should reimplement `get_releases()`.
    """
    
    def __init__(self, *args, **kwargs):
        self.current_version = None
        self.file_to_update = None
        self.release_filters = []
        super(UpdateSource, self).__init__(*args, **kwargs)


    def add_release_filter(self, filt):
        self.release_filters.append(filt)

    def test_release_filters(self, relinfo):
        """
        Returns `True` if `relinfo` should be included in the releases given the installed filters,
        otherwise `False`. Note that the platform and the version selection are not implemented by
        filters. Filters are meant to choose between different editions, or to filter out/include
        beta unstable releases.
        """
        for f in self.release_filters:
            if not f(relinfo):
                return False

        return True

    # subclasses need to reimplement:

    def get_releases(self, newer_than_version=None, **kwargs):
        """
        Should return a list of `BinReleaseInfo` of available releases. If `newer_than_version`
        argument is provided, then this function should ignore releases older or equal to the
        given argument.

        Note that for filters to work, the subclass must explicitly test each candidate release
        with `test_release_filters()`, and ignore the release if that function returns `False`.
        """
        raise NotImplementedError





# ---------------------------------------------------------------------------

class IgnoreArgument:
    pass

def _make_bin_release_info(m, lst, innerkwargs):

    logger.debug("make_bin_release_info: lst=%r", lst)

    args = {}
    for k,v in lst+innerkwargs.items():
        val = None
        if (type(v).__name__ == 'function'):
            argspec = inspect.getargspec(v)
            valargs = {}
            if ('m' in argspec.args or argspec.keywords is not None):
                valargs['m'] = m;
            if ('d' in argspec.args or argspec.keywords is not None):
                valargs['d'] = innerkwargs;
            if ('x' in argspec.args or argspec.keywords is not None):
                valargs['x'] = args; # determined args so far

            val = v(**valargs)
        else:
            val = v
            
        if (isinstance(val, IgnoreArgument)):
            continue
        
        args[k] = val

    logger.debug("make_bin_release_info: final args=%r", args)

    return BinReleaseInfo(**args)


def relpattern(re_pattern, reltype=RELTYPE_UNKNOWN, platform=None, **kwargs):
    # fix the values with default parameters
    return (re_pattern,
            (lambda m, filename, url, version=None,
             _fix_plat=platform, _fix_rtyp=reltype, _fix_kwargs=copy.deepcopy(kwargs), **innerkwargs:
             _make_bin_release_info(m,
                                    [ ('version',version) ] +
                                    _fix_kwargs.items() +
                                    [ ('filename', filename),
                                      ('url', url),
                                      ('platform',_fix_plat),
                                      ('reltype',_fix_rtyp) ],
                                    innerkwargs
                                    )
             )
            )


class ReleaseInfoFromNameStrategy(object):
    """
    Base class for a strategy to identify release details from a file name.

    Some sources need such a stategy, such as `UpdateLocalDirectorySource` and
    `UpdateGithubRelasesSource`.
    """
    def __init__(self, patterns, *args, **kwargs):
        self.patterns = [(_maybe_compile_re(r), cal) for (r, cal) in patterns]
        super(ReleaseInfoFromNameStrategy, self).__init__(*args, **kwargs)

    def get_release_info(self, filename, url, **kwargs):

        logger.debug("Trying to match filename %r to get info. kwargs=%r", filename, kwargs)
        
        for (pat,cal) in self.patterns:
            m = re.search(pat, filename)
            if m is None:
                continue

            rinfo = cal(m, filename, url, **kwargs)
            logger.debug("Got release info: %r", rinfo)
            return rinfo

        logger.warning("Can't identify info for release file named %s!" %(filename))
        return None

def _maybe_compile_re(r, flags=re.IGNORECASE):
    if (isinstance(r, type(re.compile('')))):
        return r
    return re.compile(r, flags)



def _guess_plat(m, d, default=None):
    try:
        if m.group('platform'):
            return m.group('platform').lower();
    except KeyError:
        pass

    relfile_label = d.get('relfile_label', '');
    if re.search(r'mac\s*os\s*x', relfile_label, re.I):
        return 'macosx'
    if re.search(r'linux', relfile_label, re.I):
        return 'linux'
    if re.search(r'windows', relfile_label, re.I):
        return 'win'

    if (default is not None):
        return default
    return IgnoreArgument()

def _guess_reltype(m, d, x, default=None):
    if (x.get('platform', '') == 'macosx'):
        try:
            if (m.group('onedir')):
                return RELTYPE_ARCHIVE
        except KeyError:
            pass

        return RELTYPE_BUNDLE_ARCHIVE

    if (default is not None):
        return default
    return IgnoreArgument()


_RX_VER = r'-(?P<version>\d+[\w.]+)'
_RX_VER_OPT = '('+_RX_VER+')?'
_RX_PLAT = r'-(?P<platform>macosx|linux|win)'
_RX_PLAT_OPT = '('+_RX_PLAT+')?'

_default_naming_strategy_patterns = (
#    relpattern(_RX_VER_OPT+r'-macosx\.(tar(\.gz|\.bz(ip)?2?|\.Z)|tgz|tbz2?|zip)$',
#               version=lambda m: m.group('version') if m.group('version') else IgnoreArgument(),
#               platform='macosx',
#               reltype=RELTYPE_BUNDLE_ARCHIVE),
    relpattern(_RX_VER_OPT+_RX_PLAT_OPT+r'(?P<onedir>-(onedir|dir|dist))?\.(tar(\.gz|\.bz(ip)?2?|\.Z)|tgz|tbz2?|zip)$',
               version=lambda m: m.group('version') if m.group('version') else IgnoreArgument(),
               platform=_guess_plat,
               reltype=lambda m, d, x: _guess_reltype(m, d, x, default=RELTYPE_ARCHIVE)),
    relpattern(_RX_VER_OPT+_RX_PLAT_OPT+r'\.exe$',
               version=lambda m: m.group('version') if m.group('version') else IgnoreArgument(),
               platform=lambda m, d: _guess_plat(m, d, default='win'),
               reltype=RELTYPE_EXE),
    relpattern(_RX_VER_OPT+_RX_PLAT_OPT+r'(\.(bin|run))?$',
               version=lambda m: m.group('version') if m.group('version') else IgnoreArgument(),
               platform=lambda m, d: _guess_plat(m, d, default='linux'),
               reltype=RELTYPE_EXE),
    )

# maybe e.g.
# UpdateInfoFromNameRegexpStrategy(
#     (relpattern(r'-macosx-app\.zip$', reltype=RELTYPE_BUNDLE_ARCHIVE, platform='macosx'),
#      relpattern(r'-(?P<platform>linux|win|macosx)\.zip$', reltype=RELTYPE_ARCHIVE, platform='macosx'),
#      relpattern(r'-linux.bin$', reltype=RELTYPE_EXE, platform='linux'),
#      relpattern(r'-win32.exe$', reltype=RELTYPE_EXE, platform='win'),
#     ) )
#





# -------------------------------------------------------




# simple filter for including/not including developemnt releases
class UpdateSourceDevelopmentReleasesFilter(object):
    def __init__(self, include_devel_releases=False, regexname=None):
        self.include_devel_releases = include_devel_releases

        self.regexname = regexname
        if (self.regexname is None):
            self.regexname = re.compile(r'(beta|alpha|rc)', re.IGNORECASE)

    def includeDevelReleases(self):
        return self.include_devel_releases;

    def setIncludeDevelReleases(self, include):
        self.include_devel_releases = include

    def __call__(self, relinfo):
        if (re.search(self.regexname, relinfo.get_version()) is not None):
            return self.include_devel_releases
        return True



# -------------------------------------------------------


class UpdateLocalDirectorySource(UpdateSource):
    """
    Updates will be searched for in a local directory. Useful for debugging.
    
    Will check in the given `source_directory` directory for updates. Files should be organized
    in subdirectories which should be version names, e.g.

    ::
      1.0/
        binary-macosx[.zip]
        binary-linux[.zip]
        binary-win[.exe|.zip]
      1.1/
        binary-macosx[.zip]
        binary-linux[.zip]
        binary-win[.exe|.zip]
      ...

    This updater source is mostly for debugging the module `updater4pyi`.
    """
    
    def __init__(self, source_directory, naming_strategy=None, *args, **kwargs):

        if (naming_strategy is None):
            naming_strategy = _default_naming_strategy_patterns
        if (not isinstance(naming_strategy, ReleaseInfoFromNameStrategy)):
            naming_strategy = ReleaseInfoFromNameStrategy(naming_strategy)

        self.naming_strategy = naming_strategy
        self.source_directory = source_directory

        logger.debug("source directory is %s", self.source_directory)
        
        super(UpdateLocalDirectorySource, self).__init__(*args, **kwargs)


    def get_releases(self, newer_than_version=None, **kwargs):

        try:
            versiondirs = sorted([ vdir for vdir in os.listdir(self.source_directory)
                                   if os.path.isdir(os.path.join(self.source_directory, vdir))
                                   ],
                                 key=util.parse_version,
                                 reverse=True);
        except OSError:
            logger.warning("Can't list directory %s", self.source_directory)
            raise Updater4PyiError("Can't explore directory %s" %(self.source_directory))

        logger.debug("get_releases(): Got version list: %r", versiondirs)

        newer_than_version_parsed = util.parse_version(newer_than_version)

        inf_list = []

        for ver in versiondirs:

            logger.debug("got version: %s" %(ver))
            
            if (newer_than_version_parsed is not None and
                util.parse_version(ver) <= newer_than_version_parsed):
                # no update found.
                break

            base = os.path.join(self.source_directory, ver)

            logger.debug("version %s is newer; base dir: %s" %(ver, base))

            try:
                # list files in that directory.
                for fn in os.listdir(base):
                    fnurl = util.path2url(os.path.join(base,fn))
                    logger.debug("base path %s ->  url path: %s", os.path.join(base,fn), fnurl)
                    inf = self.naming_strategy.get_release_info(filename=fn,
                                                                url=fnurl,
                                                                version=ver,
                                                                )
                    if inf is not None and self.test_release_filters(inf):
                        inf_list.append(inf)
            except OSError:
                logger.warning("Can't list directory %s", base)


        # debug: list found versions
        logger.debug("Found releases:\n"+
                     "\n".join(["\t* %s, %s (%r)" %(r.get_filename(), r.get_version(), r.__dict__)
                                for r in inf_list])
                     )
        
        # return the list of releases
        return inf_list
    







# -----------------------------------------------------------------


# github releases source



class UpdateGithubReleasesSource(UpdateSource):
    """
    Updates will be searched for in as releases of a github repo.
    """
    
    def __init__(self, github_user_repo, naming_strategy=None, *args, **kwargs):
        """
        `github_user_repo` is a string literal `'user/repo_name'`, e.g. `'phfaist/bibolamazi'`.
        """

        if (naming_strategy is None):
            naming_strategy = _default_naming_strategy_patterns
        if (not isinstance(naming_strategy, ReleaseInfoFromNameStrategy)):
            naming_strategy = ReleaseInfoFromNameStrategy(naming_strategy)

        self.naming_strategy = naming_strategy
        self.github_user_repo = github_user_repo

        super(UpdateGithubReleasesSource, self).__init__(*args, **kwargs)


    def get_releases(self, newer_than_version=None, **kwargs):

        # get repo releases.

        url = 'https://api.github.com/repos/'+self.github_user_repo+'/releases'

        try:
            fdata = upd_downloader.url_opener.open(url)
        except urllib2.URLError as e:
            logger.warning("Can't connect to github for software update check: %s", e)
            return None

        try:
            data = json.load(fdata);
        except ValueError:
            logger.warning("Unable to parse data returned by github at %s!", url)
            return None

        if (isinstance(data, dict)):
            logger.warning("Error: %s" %(data.get('message', '<no message provided>')))
            return None

        if (not isinstance(data, list)):
            logger.warning("Expected list response from github: %r", data)
            return None

        newer_than_version_parsed = None
        if (newer_than_version is not None):
            newer_than_version_parsed = util.parse_version(newer_than_version)

        inf_list = []
        
        for relinfo in data:
            html_url = relinfo.get('html_url', None)
            tag_name = relinfo.get('tag_name', None)
            rel_name = relinfo.get('name', '<unknown>')
            rel_desc = relinfo.get('body', None)
            rel_date = relinfo.get('published_at', None)
            
            # release version from tag name
            # strip starting 'v' if present
            relver = '0.0-unknown'
            if tag_name:
                relver = (tag_name[1:] if tag_name[0] == 'v' else tag_name)

            if (newer_than_version_parsed is not None and
                util.parse_version(relver) <= newer_than_version_parsed):
                logger.debug("Version %s is not strictly newer than %s, skipping...", relver, newer_than_version)
                continue
                
            relfiles = relinfo.get('assets', {})
            for relfile in relfiles:

                relfn = relfile.get('name', None)
                rellabel = relfile.get('label', None)
                relcontenttype = relfile.get('content_type', None)
                # build up the download URL
                relurl = 'https://github.com/'+self.github_user_repo+'/releases/download/'+tag_name+'/'+relfn;

                inf = self.naming_strategy.get_release_info(filename=relfn,
                                                            url=relurl,
                                                            version=relver,
                                                            # additional info:
                                                            rel_name=rel_name,
                                                            relfile_label=rellabel,
                                                            rel_description=rel_desc,
                                                            relfile_content_type=relcontenttype,
                                                            rel_tag_name=tag_name,
                                                            rel_html_url=html_url,
                                                            )
                if self.test_release_filters(inf):
                    inf_list.append(inf)

        # debug: list found versions
        logger.debug("Found releases:\n"+
                     "\n".join(["\t* %s, %s (%r)" %(r.get_filename(), r.get_version(), r.__dict__)
                                for r in inf_list])
                     )
        
        # return the list of releases
        return inf_list
    









#!/usr/bin/env python

import os
import os.path
#from distutils.core import setup
from setuptools import setup

import updater4pyi.upd_version


def read(*paths):
    """Build a file path from *paths* and return the contents."""
    with open(os.path.join(*paths), 'r') as f:
        return f.read()


setup(name='updater4pyi',
      version=updater4pyi.upd_version.version_str,
      description='Lightweight library for software auto-update for applications frozen with pyinstaller',
      long_description=read('README.md'),
      author='Philippe Faist',
      # obfuscate e-mail in source script, will be in clear in the package
      author_email=("".join([chr(ord(x)+1) for x in 'oghkhood-e`hrs?aktdvhm-bg'])),
      url='https://github.com/phfaist/updater4pyi/',
      license='BSD',
      packages=['updater4pyi'],
      package_data={'updater4pyi': [ 'cacert.pem', 'installers/unix/*.sh', 'installers/win/do_install.exe.zip' ]},
      py_modules=[],
      classifiers=[
          'Development Status :: 4 - Beta',
          'License :: OSI Approved :: BSD License',
          'Programming Language :: Python',
          'Operating System :: MacOS :: MacOS X',
          'Operating System :: Microsoft :: Windows',
          'Operating System :: POSIX :: Linux',
          'Intended Audience :: Developers',
          'Topic :: Software Development',
          'Topic :: Software Development :: Libraries :: Python Modules',
          'Topic :: System :: Software Distribution',
          ],
      )



# -*- coding: utf-8 -*-
#######################################################################################
#                                                                                     #
#   This file is part of the updater4pyi Project.                                     #
#                                                                                     #
#   Copyright (C) 2013, Philippe Faist                                                #
#   philippe.faist@bluewin.ch                                                         #
#   All rights reserved.                                                              #
#                                                                                     #
#   Redistribution and use in source and binary forms, with or without                #
#   modification, are permitted provided that the following conditions are met:       #
#                                                                                     #
#   1. Redistributions of source code must retain the above copyright notice, this    #
#      list of conditions and the following disclaimer.                               #
#   2. Redistributions in binary form must reproduce the above copyright notice,      #
#      this list of conditions and the following disclaimer in the documentation      #
#      and/or other materials provided with the distribution.                         #
#                                                                                     #
#   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND   #
#   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED     #
#   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE            #
#   DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR   #
#   ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES    #
#   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;      #
#   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND       #
#   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT        #
#   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS     #
#   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                      #
#                                                                                     #
#######################################################################################

import datetime

from PyQt4.QtCore import *
from PyQt4.QtGui import *

from . import upd_core
from . import upd_iface
from .upd_log import logger




class UpdatePyQt4Interface(QObject, upd_iface.UpdateGenericGuiInterface):

    def __init__(self, updater, parent=None, **kwargs):
        self.timer = None
        
        QObject.__init__(self, parent=parent)
        # super doesn't propagate out of the Qt multiple inheritance...
        upd_iface.UpdateGenericGuiInterface.__init__(self, updater, **kwargs)


    # ------------

    def get_settings_object(self):
        """
        Subclasses may reimplement this function to cusomize where the settings are stored.
        """
        settings = QSettings()
        settings.beginGroup('updater4pyi')
        return settings

    def load_settings(self, keylist):
        settings = self.get_settings_object()
        d = {}
        for key in keylist:
            if settings.contains(key):
                d[key] = settings.value(key).toPyObject()

        logger.debug("load_settings: read settings: %r", d)
        return d

    def save_settings(self, d=None):
        if d is None:
            d = self.all_settings()

        logger.debug("save_settings: saving settings: %r", d)

        settings = self.get_settings_object()
        for k,v in d.iteritems():
            settings.setValue(k, QVariant(v))

        # and save the settings to disk.
        settings.sync()


    def ask_first_time(self):
        msgBox = QMessageBox(parent=None)
        msgBox.setWindowModality(Qt.NonModal)
        msgBox.setText(unicode(self.tr("Would you like to regularly check for software updates%s?"))
                       %( ((unicode(self.tr(" for %s"))%self.progname)
                           if self.progname
                           else '') )
        )
        msgBox.addButton(QMessageBox.Yes)
        msgBox.addButton(QMessageBox.No)
        msgBox.setIcon(QMessageBox.Question)

        # show window as modeless window, so don't use exec_(). Instead, process application events
        # until the user has chosen some option.
        msgBox.show()
        msgBox.raise_()
        while (msgBox.isVisible()):
            QApplication.processEvents()

        # Yes, look for updates
        if (msgBox.clickedButton() == msgBox.button(QMessageBox.Yes)):
            return True

        # Anything else: no, go away
        return False


    def ask_to_update(self, rel_info):
        msgBox = QMessageBox(parent=None)
        msgBox.setWindowModality(Qt.NonModal)
        msgBox.setText(unicode(self.tr("A new software update is available (%sversion %s). "
                                       "Do you want to install it?"))
                       %(self.progname+' ' if self.progname else '', rel_info.get_version()))
        msgBox.setInformativeText(self.tr("Please make sure you save all your changes to your files "
                                          "before installing the update."))
        btnInstall = msgBox.addButton("Install", QMessageBox.AcceptRole)
        btnNotNow = msgBox.addButton("Not now", QMessageBox.RejectRole)
        btnDisableUpdates = msgBox.addButton("Disable Update Checks", QMessageBox.RejectRole)
        msgBox.setDefaultButton(btnInstall)
        msgBox.setEscapeButton(btnNotNow)
        msgBox.setIcon(QMessageBox.Question)

        # show window as modeless window, so don't use exec_(). Instead, process application events
        # until the user has chosen some option.
        msgBox.show()
        msgBox.raise_()
        while (msgBox.isVisible()):
            QApplication.processEvents()

        clickedbutton = msgBox.clickedButton()

        # btnInstall: install update
        
        if (clickedbutton == btnInstall):
            return True

        # btnNotNow, btnDisableUpdates: don't check for updates
        
        if (clickedbutton == btnDisableUpdates):
            self.setCheckForUpdatesEnabled(False)

        return False

        
    def ask_to_restart(self):
        msgBox = QMessageBox(parent=None)
        #msgBox.setWindowModality(Qt.NonModal)
        msgBox.setText(self.tr("The software update is now complete."))
        
        thisprog = str(self.tr("This program"))
        if (self.progname):
            thisprog = str('<b>'+Qt.escape(self.progname)+'</b>');
        msgBox.setInformativeText(str(self.tr("%s needs to be restarted for the changes "
                                              "to take effect. Restart now?")) % (thisprog))
        btnRestart = msgBox.addButton("Restart", QMessageBox.AcceptRole)
        btnIgnore = msgBox.addButton("Ignore", QMessageBox.RejectRole)

        msgBox.setDefaultButton(btnRestart)
        msgBox.setEscapeButton(btnIgnore)
        msgBox.setIcon(QMessageBox.Information)
        msgBox.show()
        msgBox.raise_()
        msgBox.exec_();

        clickedbutton = msgBox.clickedButton()

        if (clickedbutton == btnRestart):
            return True

        return False
        
        
    def set_timeout_check(self, interval_timedelta):

        # interval in milliseconds
        interval_ms = int(interval_timedelta.total_seconds()*1000)
        
        if self.timer is None:
            self.timer = QTimer(self)
            self.timer.timeout.connect(self.check_for_updates)
            self.timer.setSingleShot(True)
        elif self.timer.isActive():
            self.timer.stop()

        self.timer.setInterval(interval_ms)
        self.timer.start()
        logger.debug("single-shot timer started with interval=%d ms", interval_ms)


    # ------------------------------------------------------------------------------
    
    # add proper signal and slot interface / decorations to these settings
    # also, don't use timedelta, unknown to Qt4, but rather milliseconds (and the rename slot accordingly)

    
    initCheckDelayMsChanged = pyqtSignal([int])

    def initCheckDelayMs(self):
        return int(self.initCheckDelay().total_seconds()*1000)

    @pyqtSlot(int)
    def setInitCheckDelayMs(self, init_check_delay_ms, save=True):
        self.setInitCheckDelay(
            datetime.timedelta(days=0, microseconds=init_check_delay_ms*1000),
            save=save
            )

    def setInitCheckDelay(self, init_check_delay_td, save=True):
        if (self._timedelta_equal(init_check_delay_td, self.initCheckDelay())):
            # no change (up to small error)
            return
        
        super(UpdatePyQt4Interface, self).setInitCheckDelay(init_check_delay_td, save=save)
        self.initCheckDelayMsChanged.emit(int(init_check_delay_td.total_seconds()*1000))

    # ---

    checkForUpdatesEnabledChanged = pyqtSignal([bool])

    @pyqtSlot(bool)
    def setCheckForUpdatesEnabled(self, enabled, save=True):
        if (enabled == self.checkForUpdatesEnabled()):
            # no change
            return
        
        super(UpdatePyQt4Interface, self).setCheckForUpdatesEnabled(enabled, save=save)
        self.checkForUpdatesEnabledChanged.emit(enabled)

    # ---
    
    checkIntervalMsChanged = pyqtSignal([int])

    def checkIntervalMs(self):
        return int(self.checkInterval().total_seconds()*1000)

    @pyqtSlot(int)
    def setCheckIntervalMs(self, check_interval_ms, save=True):
        self.setCheckInterval(
            datetime.timedelta(days=0, microseconds=check_interval_ms*1000),
            save=save
            )

    def setCheckInterval(self, check_interval_td, save=True):
        if (self._timedelta_equal(check_interval_td, self.checkInterval())):
            # no change (up to small error)
            return
        
        super(UpdatePyQt4Interface, self).setCheckInterval(check_interval_td, save)
        self.checkIntervalMsChanged.emit(int(check_interval_td.total_seconds()*1000))

    # ---

    def _timedelta_equal(self, a, b):
        return abs( (a-b).total_seconds() ) < 1.0;
